# Copyright (c) 2021 National Technology and Engineering Solutions of Sandia, LLC.  
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering 
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this software.

# This utility converts ensemble files from one format to
# another, for example .jpg to .mp4.

# S. Martin
# 11/20/2020

# standard library imports

# command line arguments
import argparse

# logging and error handling
import logging
import sys

# miscellaneous calculations
import math

# 3rd party imports

# parallel comutation
from ipyparallel import Client

# local imports
import slypi
from slypi.ensemble import utilities

# set up argument parser
def init_parser():

    # define our own version of the slyip ensemble parser
    description = "Converts ensemble file formats.  Uses Python-like %d[::] notation, " + \
                  "where %d[::] specifies a range of numbers in a file name. For example " + \
                  '"time_step_%d[0:10:2].vtp" would specify every other file from ' + \
                  '"time_step_0.vtp" to "time_step_9.vtp".'
                  
    parser = slypi.ensemble.ArgumentParser (description=description)

    # get files to convert from command line
    parser.add_argument("--ensemble", help="Directory or directories to include in ensemble, "
                        "specified using the Python like %%d[::] notation described above.")
    parser.add_argument("--input-files", help="Files per ensemble directory to use as input "
                        "for conversion, specified using %%d[::] notation.  Note that these "
                        "files will be pre-fixed by the ensemble directories.")

    # or get files to convert from .csv file
    parser.add_argument("--csv-file", help="CSV file which specifies ensemble directories "
                        "and input files (alternate to using --ensemble and --input-files).")
    parser.add_argument("--csv-col", help="Column in CSV file where input files are "
                        "specified, can be either a string or an integer (1-based).")
    
    # input options
    parser.add_argument("--input-format", help="Format for input files (optional, inferred "
                        "from file extension if not provided).")

    # output arguments and options
    parser.add_argument("--output-dir", help="Directory to place output.  All files will be "
                        "stored using directories that mirror those specified by --ensemble.")
    parser.add_argument("--output-format", help="File output format information, "
                        "e.g. file extension.")
    parser.add_argument("--over-write", action="store_true", help="Over-write output "
                        "directory if already present.")

    # output a csv file
    parser.add_argument("--csv-out", help="File name of output .csv file with file links for "
                        "converted files (optional).  Will be written to output directory.")
    parser.add_argument("--csv-header", help="Name of output files header, needed only "
                        "if writing out a .csv file.")

    # parallel option using ipyparallel
    parser.add_argument('--parallel', default=False, action="store_true", 
                        help="Use ipyparallel (must be available and running).")

    return parser

# check arguments
def check_arguments(log, args):

    # check for command line input arguments
    if args.csv_file is None:

        # make sure the ensemble argument is present
        if args.ensemble is None:
            log.error("Ensemble directories are required.  Please use --ensemble and try again.")
            sys.exit(1)
        
        # make sure the input argument is present
        if args.input_files is None:
            log.error("Input files are required.  Please use --input-files and try again.")
            sys.exit(1)

        # disable args.csv_col (if provided)
        args.csv_col = None

    # otherwise check that ensemble arguments are not present
    else:

        if args.ensemble is not None or args.input_files is not None:
            log.error("Please specify either --csv-file or --ensemble, but not both.")
            sys.exit(1)

        # also check that csv column is provided
        if args.csv_col is None:
            log.error("Please specify --csv-col to provide input files.")
            sys.exit(1)

    # make sure the output directory is present
    if args.output_dir is None:
        log.error("Output directory must be specified.  Please use --output-dir and try again.")
        sys.exit(1)
    
    # make sure the output format is specified
    if args.output_format is None:
        log.error("Output format is required.  Please use --output-format and try again.")
        sys.exit(1)

    # if csv is to be written out, must also have column header
    if args.csv_out is not None:
        if args.csv_header is None:
            log.error("CSV header is required to output .csv file.  " +
                      "Please use --csv-header and try again.")
            sys.exit(1)

# convert ensemble, either in serial or parallel
def convert_ensemble (plugin, log ,args, ensemble_table, ensemble_dirs, mirror_dirs):
    
    num_ensemble_dirs = len(ensemble_dirs)

    if not args.parallel:

        # go through each directory and convert files
        files_written = []
        for i in range(num_ensemble_dirs):

            # convert a single simulation
            files_to_convert, files_created, files_converted = convert_simulation(plugin, args,
                ensemble_table, ensemble_dirs[i], mirror_dirs[i])

            # log progress
            progress_report(log, files_to_convert, ensemble_dirs[i], files_created)

            # keep track of files written in csv specifier format
            files_written.append(files_converted)

        return files_written
    
    # parallel operation, using direct view
    rc = Client()

    # use cloudpickle because pickle won't work with plugins
    rc[:].use_cloudpickle()

    # get number of available engines
    num_engines = len(rc)
    
    # push out parallel jobs, one load per engine
    files_written = []
    for i in range(0, num_ensemble_dirs, num_engines):

        # get block of simuilations
        block_dirs = [ensemble_dirs[j] 
            for j in range(i, i + num_engines) if j < num_ensemble_dirs]
        mirror_block = [mirror_dirs[j] 
            for j in range(i, i + num_engines) if j < num_ensemble_dirs]

        # push out jobs per engine
        async_results = []
        for j in range(len(block_dirs)):
            async_result = rc[j].apply_async(convert_simulation, 
                plugin, args, ensemble_table, block_dirs[j], mirror_block[j])
            async_results.append(async_result)

        # wait for results
        rc[:].wait(async_results)

        # get results
        for j in range(len(async_results)):

            # get items in order they were put into queue
            async_result = async_results.pop(0)

            # get returned values
            files_to_convert = async_result.get()[0]
            files_created = async_result.get()[1]
            files_converted = async_result.get()[2]

            # log progress
            progress_report(log, files_to_convert, block_dirs[j], files_created)

            # save csv results
            files_written.append(files_converted)

        # clean up ipyparallel
        rc.purge_everything()

    # clean up sockets
    rc.close()
    
    return files_written

# convert a single simulation (can be run in parallel)
def convert_simulation (plugin, args, ensemble_table, ensemble_dir, mirror_dir):

    # find files in directory
    files_to_convert = ensemble_table.files(ensemble_dir)

    # no files, then return empty
    if files_to_convert == []:
        return [], []

    # convert files
    files_created = plugin.convert_files(files_to_convert, mirror_dir, 
        args.output_format, input_type=args.input_format)

    # if only one file created, use the file name
    if len(files_created) == 1:
        files_written = files_created[0]

    # convert input file specifier to output file specifier
    else:
        files_written = ensemble_table.convert_specifier(ensemble_dir,
            args.output_dir, args.output_format)

    return files_to_convert, files_created, files_written

# progress report to log for conversion
def progress_report(log, files_to_convert, ensemble_dir, files_created):

    # print error if no files were converted
    if files_to_convert == []:
        log.error("No files to convert, please provide existing files for input.")
        sys.exit(1)

    # report on files converted
    log.info("Found %d file(s) in ensemble directory %s." % 
        (len(files_to_convert), ensemble_dir))
    for j in range(len(files_created)):
        log.info('File written "%s".' % files_created[j])

# converts ensemble files from one format to another
# call programmatically using arg_list
def main(arg_list=None):

    # initialize parser
    parser = init_parser()

    # parse arguments
    if arg_list is not None:

        # parse in case called with arg_list
        args, arg_list = parser.parse_args(arg_list)

    else:

        # command line parser
        args, arg_list = parser.parse_args()

    # start logger
    slypi.ensemble.init_logger(log_file=args.log_file, log_level=args.log_level)
    log = logging.getLogger("slypi.ensemble.convert")
    log.debug("Started convert.")

    # check arguments
    check_arguments(log, args)

    # import and initialize plugin
    plugin, unknown_args = slypi.ensemble.plugin(args.plugin, arg_list)

    # check for un-recognized arguments
    if len(unknown_args) > 0:
        log.error("Unrecognized arguments: %s.  Please try again." % str(unknown_args))
        sys.exit(1)

    # create ensemble table
    ensemble_table = utilities.Table(log, csv_file=args.csv_file,
        ensemble_spec=args.ensemble, file_spec=args.input_files, 
        header="convert")

    # get ensemble directories/file specs
    if args.csv_file is not None:
        ensemble_dirs = ensemble_table.get_col(args.csv_col)
    else:
        ensemble_dirs = ensemble_table.get_col("convert")

    # quit if no directories found
    num_ensemble_dirs = len(ensemble_dirs)
    if num_ensemble_dirs == 0:
        log.error("No ensemble directories found.  " +
                  "Please identify existing directories and try again.")
        sys.exit(1)
    else:
        log.info("Found %d ensemble directory(ies)." % num_ensemble_dirs)

    # mirror directories to output direoctory
    mirror_dirs = ensemble_table.mirror_directories(
        args.output_dir, ensemble_dirs, args.over_write)
    
    # if directory already exists, quit
    if mirror_dirs is None:
        log.error("Output directory already exists, use --over-write if " +
            "you want to over-write or add to previous contents.")
        sys.exit(1)
    log.info("Output directory created or already exists with mirrored file structure.")

    # convert files, either in serial or parallel
    files_written = convert_ensemble(plugin, log, args, ensemble_table, 
        ensemble_dirs, mirror_dirs)

    # are we writing out a csv file?
    if args.csv_out is not None:

        # add new column to ensemble table
        ensemble_table.add_col(files_written, args.csv_header)

        # put .csv file in output directory
        ensemble_table.to_csv(args.csv_out, output_dir=args.output_dir, 
            cols=[args.csv_header])

# entry point for command line call
if __name__ == "__main__":
    main()


# Copyright (c) 2021 National Technology and Engineering Solutions of Sandia, LLC.  
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering 
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this software.

# This module contains utility funtions for use in processing 
# ensemble data.

# S. Martin
# 11/25/2020

# standard library
import re
import os
import posixpath

# 3rd party libraries
import numpy as np
import pandas as pd
from ipyparallel import Client

# class for keeping track of ensemble files
class Table:
    """
    Provides storage and methods for keeping track of simulation data and
    files in an ensemble.  The central assumption is that the esnemble is
    stored in a directory structure of the form (names are arbitrary):

    ::

        ensemble
        |-- ensemble.info
        |-- simulation.1
            |-- simulation.info
            |-- time.step.1
            |-- time.step.2
            |-- ...
        |-- simulation.2
        |-- simulation.3
        ...

    where ensemble is the central directory, containing multiple simulations,
    each having it's own directory simulation.1, simulation.2, and so on.  The
    simulation directories then contain files containg time step data time.step.1,
    time.step.2, etc.  The directory names can be somewhat arbitrary and can 
    have additional subdirectories and files, but the convention for using
    these utitilities is that the simulation directories can be specified using
    a Python like ``%d[::]`` specifier.  For example, for the above ensemble,
    we would use ``ensemble/simulation.%d[1:]`` to specify the file name 
    format for each of the simulation folders, then we would use
    ``time.step.%d[1:]`` to specify the time step files within each simulation
    folder.

    The ``%d[::]`` notation specifies the order and numbers in the directory/file
    names and using the Python slicing conventions.  For example ``%d`` specifies
    all names with an integer in the given location, ``%d[5:10]`` specifies
    all names with an integer starting at 5 and ending at 9, and ``%d[100:2:-2]``
    specifies all names with an integer starting at 100 and descending by 2 to 2.

    The numbers in the directory/file names are assumed to be >= 0 but are otherwise
    unrestriced.

    To instantiate an ensemble.Table object, use either a .csv file to
    create a full table, or an ensemble_spec, file_spec, and table header to
    create a single column table.

    Args:
        csv_file=None (string): file name of .csv file
        ensemble_spec=None (string): string with %d[::] giving 
            simulation directories
        file_spec=None (string): string with %d[::] giving time step file names
        header=None (string): name for column header
    
    Note: If csv_file is provided then the other inputs are ignored.  If no
    inputs are provided then an empty table is created.
    """

    # the ensemble information is stored internally using all available
    # information, for example full file names.
    def __init__(self, log, data_frame=None, csv_file=None, ensemble_spec=None, 
                 file_spec=None, header=None):

        # logger to use
        self.log = log

        # initialize dataframe to empty
        self.table = pd.DataFrame()

        # data_frame over-rides all other options
        if data_frame is not None:

            self.table = data_frame

        # csv file over-rides ensemble/file/header spec
        elif csv_file is not None:

            # load csv file into data frame
            self.table = pd.read_csv(csv_file, index_col=0)

        # otherwise use ensemble specifier
        elif ensemble_spec is not None:

            # get ensemble directories (in order)
            ensemble_dirs = self.directories(ensemble_spec)

            # combine with ensemble directories to create table
            table_col = []
            for ensemble_dir in ensemble_dirs:
                if file_spec is not None:
                    table_col.append(os.path.join(ensemble_dir, file_spec))
                else:
                    table_col.append(ensemble_dir)
            
            if header is not None:
                self.table = pd.DataFrame(data={header: table_col}, index=ensemble_dirs)
            else:
                self.table = pd.DataFrame(data={'': table_col}, index=ensemble_dirs)

        log.debug('Created ensemble table.')

    # parses a file or directory name according to %d[::] format
    # and returns the associated integer, if present, otherwise returns None
    def _parse_d_name(self, file_or_dir, root, ext):

        # skip non-root matches
        if file_or_dir.find(root) != 0:
            return None

        # skip non-extension matches
        if not file_or_dir.endswith(ext):
            return None

        # make sure dir_name can contain both root and extension
        if len(file_or_dir) < len(root) + len(ext):
            return None
        
        # strip off root and extension
        file_or_dir_num = file_or_dir[len(root) : len(file_or_dir) - len(ext)]

        # check that dir_num is an integer
        try:
            file_or_dir_num = int(file_or_dir_num)
        except ValueError:
            return None
        
        return file_or_dir_num

    # finds and orders all files or directories in a path matching specifier
    # use is_dir == True to list directories, False to list files
    # root, start, stop, step, and ext specify the %d[::] format,
    # and path_ext supplies any trailing sub-directories (in case is_dir==True)
    def _catalog_path_contents(self, path, is_dir, root, 
        start, stop, step, ext, path_ext=""):

        # get contents of directory
        path_contents = os.listdir(path)
        
        # catalog directories matching %d[::] format
        catalog_name = []
        catalog_num = []
        for file_or_dir in path_contents:

            # skip files
            if is_dir:
                if not os.path.isdir(os.path.join(path, file_or_dir)):
                    continue

            # skip directories
            else:
                if not os.path.isfile(os.path.join(path, file_or_dir)):
                    continue
                    
            # get associated number, if any
            file_or_dir_num = self._parse_d_name(file_or_dir, root, ext)

            # if successful match, return name and number
            if file_or_dir_num is not None:
                catalog_name.append(file_or_dir)
                catalog_num.append(file_or_dir_num)

        # check that files were found
        if len(catalog_name) == 0:
            return []

        # if start is not given use min or max
        if start is None:

            # if step is positive, start at min
            if step > 0:
                start = min(catalog_num)

            # if step is negative, start at max
            if step < 0:
                start = max(catalog_num)

        # if stop is not given use min or max
        if stop is None:

            # if step is positive, stop at max (including last file)
            if step > 0:
                stop = max(catalog_num) + 1

            # if step is negative, stop at min (including last file)
            if step < 0:
                stop = min(catalog_num) - 1

        # precompute range of acceptable numbers
        range_nums = range(start, stop, step)

        # filter according to range
        for i in reversed(range(len(catalog_num))):

            # check if number is in range
            if catalog_num[i] not in range_nums:
                del catalog_name[i]
                del catalog_num[i]

        # sort ascending or descending depending on step sign
        sort_inds = np.argsort(catalog_num)
        if step < 0:
            catalog_name = list(np.array(catalog_name)[sort_inds[::-1]])
        else:
            catalog_name = list(np.array(catalog_name)[sort_inds])
        
        # add directory information to names, and double check
        # that files/directories exist
        sorted_catalog_names = []
        for file_or_dir in catalog_name:

            # add directories
            if is_dir:
                if os.path.exists(os.path.join(path, file_or_dir, path_ext)):
                    sorted_catalog_names.append(os.path.join(path, file_or_dir, path_ext))

            # add files (including path)
            else:
                if os.path.exists(os.path.join(path, file_or_dir)):
                    sorted_catalog_names.append(os.path.join(path, file_or_dir))

        return sorted_catalog_names

    # returns a list of directories matching %d[::] specifier
    def directories(self, directory_spec):
        """
        Return a list of directories matching ``%d[::]`` specifier.  The specifier
        is expanded and existing directories are identified and returned.

        Args:
            directory_spec (string): directory name with ``%d[::]`` specifier

        Returns:
            directory_list (list): list of directories matching specifier
        """

        # parse directory spec, note that if %d is in middle of multiple 
        # directories then dir_ext could include other directories
        dir_root, start, stop, step, dir_ext = \
            parse_d_format(self.log, directory_spec)

        # in case %d was not given, check that directory exists
        if step is None:
            if os.path.isdir(directory_spec):
                return [directory_spec]
            else:
                return []

        # parse directory specifier into parts path/root%d[::]ext/path_ext

        # get path of directories and root of spec
        path, root = os.path.split(dir_root)

        # get directory spec and path extensions
        path_ext, ext = os.path.split(dir_ext[::-1])
        ext = ext[::-1]
        path_ext = path_ext[::-1]

        return self._catalog_path_contents(path, True, 
            root, start, stop, step, ext, path_ext=path_ext)

    # creates output directories, return false if already exist
    def mirror_directories (self, output_dir, ensemble_dirs, over_write):
        """
        Creates a set of directories in the output directory which mirror
        the ensemble directory structure, unless output directory already
        exists.

        Args:
            output_dir (string): name of output directory to create
            ensemble_dirs (list): list of ensemble directories to mirror
            over_write (boolean): true to over-write existing directories

        Returns:
            mirror_dirs (list): list of mirror directories (including output_dir),
                None if directories were not created
        """

        # get common path
        if len(ensemble_dirs) > 1:
            common_path = os.path.commonpath(ensemble_dirs)

        # if only one simulation, commom path is everything up to file specifier
        else:
            common_path = os.path.dirname(ensemble_dirs[0])

        # construct mirror directories
        mirror_dirs = []
        for i in range(len(ensemble_dirs)):

            # strip off file
            ensemble_path = os.path.dirname(ensemble_dirs[i])

            # remove common path
            mirror_path = ensemble_path.split(common_path)[1]
            mirror_path = os.path.basename(mirror_path)

            # strip off base dir
            mirror_dirs.append(os.path.join(output_dir, mirror_path))

        # check if output directory exists
        if os.path.exists(output_dir):
            if not over_write:
                return None
        
        # create mirror directories
        for mirror_dir in mirror_dirs:
            try:
                os.makedirs(mirror_dir)
            except FileExistsError:

                # directory already exists, but that's OK because
                # user has allowed over-writing
                pass

        return mirror_dirs
    
    # locate and order existing files in ensemble directory matching %d[::] specifier
    def files (self, file_spec):
        """
        Return a list of files matching ``%d[::]`` specifier.  The specifier
        is expanded and existing files are identified and returned.

        Args:
            file_spec (string): file path with ``%d[::]`` specifier

        Returns:
            file_list (list): list of files matching specifier
        """

        # split into path and specifier
        path, files = os.path.split(file_spec)

        # parse file spec
        root, start, stop, step, ext = parse_d_format(self.log, files)

        # in case %d was not given, check that file exists
        if step is None:
            if os.path.isfile(file_spec):
                return [file_spec]
            else:
                return []

        return self._catalog_path_contents(path, False, 
            root, start, stop, step, ext)
    
    # get all files in an ensemble
    def ensemble_files (self, ensemble_dirs, parallel=False):
        """
        Returns a list of lists of files matching ``%d[::]`` specifier in
        the ensemble_dirs.  The specifier is expanded and existing files are
        identified and returned.  The directories are expected to exist.

        Args:
            ensemble_dirs (list): list of ensemble directories to expand (with specifier)
            parallel (boolean): run in parallel with ipyparallel (default False)

        Returns:
            sim_files (list): list of list of files matching specifier
        """

        # if not parallel, just get all sim files normally
        if not parallel:
            return self._get_sim_files(ensemble_dirs)

        # otherwise start parallel version

        # parallel operation, using direct view
        rc = Client()
        view = rc.load_balanced_view()

        # get number of available engines
        num_engines = len(rc)

        # break files into blocks so that we use all available engine
        num_dirs = len(ensemble_dirs)
        block_size = int(np.ceil(num_dirs/num_engines))

        # push out parallel jobs finding ensemble files
        async_results = []
        for i in range(num_engines):

            # get block of ensemble directories to expand
            block_dirs = [ensemble_dirs[j] 
                for j in range(i*block_size, i*block_size + block_size) if j < num_dirs]

            # expand ensemble directories as nodes are available
            async_result = rc[i].apply_async(self._get_sim_files, block_dirs)
            async_results.append(async_result)
        
        # wait for calculation to complete
        view.wait(async_results)

        # collect results
        sim_files = []
        for i in range(len(async_results)):

            # get items in order they were put into queue
            async_result = async_results.pop(0)

            # retrieve data from result
            sim_files += async_result.get()

        # clean up ipyparallel
        rc.purge_everything()

        # close ZMQ sockets
        rc.close()

        return sim_files

    # get simulation files for a block of ensemble directories
    def _get_sim_files (self, ensemble_dirs):

        sim_files = []
        for i in range(len(ensemble_dirs)):

            # find files in directory
            files_to_convert = self.files(ensemble_dirs[i])

            # add files to list
            sim_files.append(files_to_convert)

        return sim_files

    # convert an input file specifier to an output file specifier
    def convert_specifier (self, file_spec, output_dir, output_type):
        """
        Returns an output file specifier matching the provided input specifier.

        Args:
            file_spec (string): file path with ``%d[::]`` specifier
            output_dir (string): output directory for file spec files
            output_type (string): file extension of output files

        Returns:
            out_file_spec (string): output file path with ``%d[::]`` specifier
        """

        # create output file specifier
        file_name = os.path.basename(file_spec)
        file_root, file_ext = os.path.splitext(file_name)
        file_spec_out = os.path.join(output_dir, file_root + "." + output_type)

        return file_spec_out

    # get a column from the table
    def get_col (self, col):
        """
        Returns a column from the ensemble table.

        Args:
            col (string): name of column to return
        
        Returns:
            col_list (list): list of contents in column        
        """

        return self.table[col].tolist()

    # add a column to the table
    def add_col (self, col, header):
        """
        Adds a column to the ensemble table.  This column can be a list or a dictionary.
        If it is a list, the column is added in the list order.  If it is a dictionary,
        the column is added in the order of the dictionary keys by matching with an
        existing column.  The column is added at the end of the table.

        Args:
            col (list or dict): column data to add
            header (string): name of new column
        """

        self.table[header] = col

    # convert file pointers in specified columns
    def convert_cols (self, cols, uri_root):
        """
        Converts the specified columns in the table to have the given URI root.

        Args:
            cols (list): names of columns to convert
            uri_root (string): URI root to use for conversion

        Note: Resulting file is output in unix format (forward slashes).
        """

        # check that columns exist in table
        self._check_cols(cols)

        # go through each column and convert file paths
        for col in cols:
            
            # get column
            old_col = self.table[col].to_list()

            # convert column
            new_col = self._convert_col_uri(old_col, uri_root)

            # replace column in dataframe
            self.table[col] = new_col

    # converts a list of paths to a URI based path
    def _convert_col_uri (self, paths, uri_root):

        # get common path
        common_path = os.path.commonpath(paths)

        # go through list
        new_paths = []
        for path in paths:

            # remove common path
            new_path = path.split(common_path)[1]

            # reverse path to get rid of any leading "/" which
            # might be interpreted as a root directory
            dir_path = os.path.basename(new_path[::-1])[::-1]
            base_path = os.path.dirname(new_path[::-1])[::-1]
            new_path = os.path.join(dir_path, base_path)

            # convert to unix, just in case
            new_path = new_path.replace ('\\','/')

            # add URI root
            uri_path = posixpath.join(uri_root, new_path)

            # save result in list of new paths
            new_paths.append(uri_path)

        return new_paths

    # check that provided columns exist in table
    # this prevents an ugly and uninterpretable pandas error
    def _check_cols (self, cols):

        # make sure columns are specified
        if cols is None:
            return

        # check that columns exist in table
        for col in cols:            
            if col not in self.table.columns:
                raise ValueError('Could not find column "%s" in table.' % col)

    # write out .csv file
    def to_csv (self, file_out, output_dir='', cols=None, exc_cols=None, 
                index=True, index_label=None):
        """
        Writes out the table to a .csv file.

        Args:
            file_out (string): name of .csv file
            output_dir (string): output directory to use for .csv file
            cols (list): list of column headers to output
            exc_cols (list): list of column headers to exclude from output
            index (boolean): write out index column
            index_label (string): use as index header
        """

        # check that columns exist in table
        self._check_cols(cols)

        # check that excluded columns exist
        self._check_cols(exc_cols)

        # generate all columns to include
        if cols is None:
            cols = list(self.table.columns)
        if exc_cols is not None:
            for i in reversed(range(len(cols))):
                if cols[i] in exc_cols:
                    cols.pop(i)

        # check if there are any columns to write
        if len(cols) == 0:
            raise ValueError('All columns have been excluded -- nothing to write for file "%s".' %
                             file_out)

        # put output file into output_dir
        csv_out_file = os.path.join(output_dir, file_out)

        self.table.to_csv(path_or_buf=csv_out_file, columns=cols, 
            index=index, index_label=index_label)
        self.log.info('File written: %s.' % csv_out_file)


# helper function for parsing %d[::] format string
def parse_d_format(log, d_str):
    """
    Parse %d[::] format into root, start, stop, step, and extension,

    where root is the string before the %d[], start, stop, step are
    integers to put into range(start, stop, step), and extension is 
    the string after the %d[::].

    if %d is not given, start, stop, step are None, and ext is ''.
    if %d is given, but [] is empty, start=None, stop=None, step=1.
    if %d is given as [::], start=None, stop=None, step=1.
    if %d is given as [::-1], start=None, stop=None, step=-1.
    if %d is given as [:stop] then start=0, stop=None, step=1.

    In other words, step==None only if %d is not given, in all other
    cases, step is non-zero, but start, stop might have to be inferred.

    Args:
        (object) log: logger for messages
        (string) d_str: string to parse

    Returns:
        root (string): string before %d
        start (int): start index
        stop (int): stop index
        step (int): step between start and stop
        ext (string): extension after %d[::]
    """

    # get root
    d_str_parts = d_str.split("%d")

    # if there is no %d return original string
    if len(d_str_parts) == 1:
        return d_str, None, None, None, ''

    # if there is more than one %d raise exception
    if len(d_str_parts) > 2:
        log.error('Invalid %d[::] format, ' +
            'can not have more than one "%%d": %s' % d_str)
        raise EnsembleSpecifierError(d_str, 'can not have more than one "%d"')

    # seperate root, possibly extension
    root = d_str_parts[0]
    ext = d_str_parts[1]

    # split on [ after %d
    left_brackets = ext.split("[")

    # if no brackets, return empty range list
    if len(left_brackets) == 1:
        return root, 0, None, 1, ext

    # if more than one left bracket then error
    if len(left_brackets) > 2:
        log.error('Invalid %d[::] format, ' +
            'can not have more that one "[": %s' % d_str)
        raise EnsembleSpecifierError(d_str, 'can not have more than one "["')

    # check that left bracket is after %d
    if left_brackets[0] != '':
        log.error("Invalid %%d[::] format: %s." % d_str)
        raise EnsembleSpecifierError(d_str, "invalid %d[::] format")

    # split on ] after [
    right_brackets = left_brackets[1].split("]")

    # error if no closing bracket
    if len(right_brackets) == 1:
        log.error('Invalid %%d[::] format, has no closing "]": %s' % d_str)
        raise EnsembleSpecifierError(d_str, 'has no closing "]"')

    # error if more than one right bracket
    if len(right_brackets) > 2:
        log.error('Invalid %d[::] format, ' +
            'can not have more than one "]": %s' % d_str)
        raise EnsembleSpecifierError(d_str, 'can not have more than one "]"')

    # save extension, what's between the brackets
    ext = right_brackets[1]
    in_brackets = right_brackets[0]

    # check that between brackets has only : or digits
    if len(re.findall('[^:\-\d]', in_brackets)) > 0:
        log.error('Invalid %d[::] format, ' +
            'can only have integers between "[]": %s' % d_str)
        raise EnsembleSpecifierError(d_str, 'can only have integers between "[]"')

    # get numbers between brackets
    nums_in_brackets = in_brackets.split(":")

    # check that there are at least two numbers
    if len(nums_in_brackets) < 2 or len(nums_in_brackets) > 3:
        log.error('Invalid %d[::] format, ' +
            'must have at two or three numbers between "[]": %s' % d_str)
        raise EnsembleSpecifierError(d_str, 'must have two or three numbers between "[]"')

    # convert numbers to start, stop, step
    # unspecified numbers are returned as None
    start = None
    stop = None
    step = 1
    try:

        # convert start
        if nums_in_brackets[0] != '':
            start = int(nums_in_brackets[0])

        # convert stop
        if nums_in_brackets[1] != '':
            stop = int(nums_in_brackets[1])

        # convert step
        if len(nums_in_brackets) == 3:
            if nums_in_brackets[2] != '':
                step = int(nums_in_brackets[2])
    
    except ValueError:
        log.error('Invalid %d[::] format, ' +
            'must have integer values between "[]": %s' % d_str)
        raise EnsembleSpecifierError(d_str, 'must have integer values between "[]"')

    # check value for start is non-negative
    if start is not None:
        if start < 0:
            log.error('Invalid %d[::] format, ' +
                'must have non-negative start between "[]": %s' % d_str)
            raise EnsembleSpecifierError(d_str, 'must have non-negative start between "[]"')

    # check value for stop is non-negative
    if stop is not None:
        if stop < 0:
            log.error('Invalid %d[::] format, ' +
                'must have non-negative stop between [""]: %s' % d_str)
            raise EnsembleSpecifierError(d_str, 'must have non-negative stop between "[]"')
    
    # check that step is not zero
    if step == 0:
        log.error('Invalid %d[::] format, ' +
            'must have non-zero step between [""]: %s' % d_str)
        raise EnsembleSpecifierError(d_str, 'must have non-zero step between "[]"')

    return root, start, stop, step, ext


# factory method to combine tables
def combine (log, tables, ignore_index=False):
    """
    Combines multiple tables to create a new table.  Tables are assumed
    to have identical index values and are joined as columns.

    Args:
        log (logger object): logger for writing output
        tables (list): list of ensemble tables to combine
        ignore_index (bool): ignore index values and join as rows

    Returns:
        combined_table (Table): ensemble table made up of input tables
    """
    
    # if no tables are given, return empty table
    if len(tables) == 0:
        return Table(log)
    
    # if one table is given, return table
    if len(tables) == 1:
        return tables[0]
    
    pd_tables = []
    for table in tables:

        # check that indices are identical
        if not ignore_index:
            if not np.array_equal(table.table.index.to_numpy(), 
                                  tables[0].table.index.to_numpy()):
                log.error("Indices do not match -- tables cannot be combined. " +
                                 "Use --ignore-index to combine anyway.")
                raise ValueError("Indices do not match -- tables cannot be combined. " +
                                 "Use --ignore-index to combine anyway.")

            # use original indices
            reset_table = table.table

        # otherwise reset indices
        else:
            reset_table = table.table.reset_index(drop=True)
    
        # remove Table information, leaving only pandas table
        pd_tables.append(reset_table)
        
    # concatenate tables
    return Table(log, data_frame=pd.concat(pd_tables, axis=1, join='inner', verify_integrity=True))


# factory method to explode a table using a column of lists
def explode(log, table, table_col, table_list):
    """
    Expand a table by replacing a given col with an exploded column,
    where other rows are duplicated.

    Args:
        log (logger object): logger for writing output
        table (Table object): ensemble table to explode
        table_col (string or integer): column in table to replace with exploded column
        table_list (list): column with list items to explode to replace table_col
    """

    # replace table_col with table_list
    new_col = pd.DataFrame({table_col: table_list},
        index=table.table.index)
    table.table[table_col] = new_col[table_col]
    
    # explode table using table_list column
    return Table(log, data_frame=table.table.explode(table_col))

# factory method to concatenate tables
def concat(log, tables):
    """
    Concatenates multiple tables to create a new table.  Tables are assumed
    to have identical header values and are joined vertically.

    Args:
        log (logger object): logger for writing output
        tables (list): list of ensemble tables to combine

    Returns:
        concat_table (Table): ensemble table made up of input tables
    """

    # if no tables are given, return empty table
    if len(tables) == 0:
        return Table(log)
    
    # if one table is given, return table
    if len(tables) == 1:
        return tables[0]
    
    # check that column headers are identical
    headers = list(tables[0].table)
    headers_identical = True
    for table in tables:
        if list(table.table) != headers:
            headers_identical = False

    # quit if headers are not identical
    if not headers_identical:
        log.error("Table headers are not identical, cannot concatenate tables.")
        raise ValueError ("Table headers are not identical, cannot concatenate tables.")

    # make list of tables
    pd_tables = []
    for table in tables:
        pd_tables.append(table.table)

    return Table(log, data_frame=pd.concat(pd_tables))


# error message for incorrect %d[::] specifier
class EnsembleSpecifierError(Exception):
    """
    Exception raised for errors in %d[::] format specifier.

    Args:
        specifier (string): input specifier which caused the error
        message (string): explanation of the error
    """

    def __init__(self, specifier, message="invalid %d[::] format"):

        # specifics for error
        self.specifier = specifier
        self.message = message

        # raise exception
        super().__init__(self.message)

    def __str__(self):
        return f'{self.message}: {self.specifier}'



# Copyright (c) 2021 National Technology and Engineering Solutions of Sandia, LLC.  
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering 
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this software.

# This script contains for code manipulating ensemble .csv files, including 
# creation from input decks, joining multiple tables, and expanding
# a table according to rules provided by plugin.

# S. Martin
# 3/23/2021

# standard library imports
import argparse
import logging
import sys
import os

# 3rd party imports

# local imports
import slypi.ensemble as ensemble
from slypi.ensemble.utilities import Table as EnsembleTable
from slypi.ensemble.utilities import combine as ensemble_combine

# set up argument parser
def init_parser():

    # define our own version of the slypi.ensemble parser
    description = "Manipulates .csv files from ensemble data."
    parser = ensemble.ArgumentParser (description = description)

    # major csv options (must choose one)
    parser.add_argument('--create', action="store_true", help="Create ensemble .csv file "
                        "from simulation input decks.")
    parser.add_argument("--join", nargs="+", help="List of slypi.ensemble .csv files to join horizontally " +
                        "(first column is assumed to be index).")
    parser.add_argument("--concat", nargs="+", help="List of slypi.ensemble .csv files to join vertically " +
                        "(all column headers must be identical).")
    parser.add_argument('--expand', help='Expand links in .csv file to include data in table.  Uses '
                        'plugin to expand links.')

    # create/join csv from ensemble specifier and input files
    parser.add_argument("--ensemble", help="Directory or directories to include in ensemble, "
                        "specified using the Python like %%d[::] notation described above.")
    parser.add_argument("--input-files", help="Files per ensemble directory to use as input "
                        "for metadata, specified using %%d[::] notation.  Note that these "
                        "files will be pre-fixed by the ensemble directories.")
    parser.add_argument("--input-header", help='Name to assign input file header, e.g. "Input Deck"')

    # create specific option
    parser.add_argument("--input-format", help="Format for input files.  Optional, inferred "
                        "from file extension if not provided.")

    # join specific options
    parser.add_argument("--ignore-index", action="store_true", default=False, help="Ignore "
                        "index column when joining tables.")
    parser.add_argument("--convert-cols", nargs="+", help="Converts the given columns using " +
                        "--uri-root-out when joining tables.")
    parser.add_argument("--uri-root-out", help="Root name of URI used to transform file " +
                        "pointers in .csv output file when joining files.  Note that this " +
                        "will only work if the file pointers have a common root.  All paths in " +
                        "this case will be converted to unix paths.")

    # concat specific options
    parser.add_argument("--add-origin-col", help="Add a column containing the data origin. " +
                        "This flag gives the new column name.")
    parser.add_argument("--origin-col-names", nargs="+", help="Names to use for origin column, " +
                        "one per file to concatenate (defaults to file names).")

    # expand specific options
    parser.add_argument("--expand-header", help="Table column to expand (either name or index).")

    # output file information
    parser.add_argument("--output-dir", help="Output directory for any files produced.")
    parser.add_argument("--csv-out", help="File name of output .csv file.")
    parser.add_argument("--csv-no-index", action="store_false", default=True,
                        help="Do not output the index column.")
    parser.add_argument("--csv-index-header", default=None, help="Index header name for .csv file " +
                        "(default is None).")
    parser.add_argument("--csv-headers", nargs='*', help="Output only the given headers to " +
                        "the .csv file (defaults to all headers).")
    parser.add_argument("--exclude-csv-headers", nargs='*', help="Exclude the given headers from " +
                        "the .csv file.")

    # over-write if file exists
    parser.add_argument("--over-write", action="store_true", help="Over-write output "
                        "file if already present.")

    return parser

# check arguments for create option
def check_create_arguments(log, args):

    # make sure the ensemble argument is present
    if args.ensemble is None:
        log.error("Ensemble directories are required.  Please use --ensemble and try again.")
        sys.exit(1)
    
    # make sure the input argument is present
    if args.input_files is None:
        log.error("Input files are required.  Please use --input-files and try again.")
        sys.exit(1)

    # make sure the input file header is present
    if args.input_header is None:
        log.error("Input header is required.  Please use --input-header and try again.")
        sys.exit(1)

# check arguments for join
def check_join_arguments(log, args):
    
    # if only one csv file, must specify ensemble parameters
    if len(args.join) == 1:

        if args.ensemble is None or \
           args.input_files is None or \
           args.input_header is None:

            log.error("If only using one .csv file, you must specify --ensemble arguments.  " +
                      "Please use --ensemble, --input-files, --input-header and try again. ")
            sys.exit(1)
    
    # if ignoring index column you can't output index
    if args.ignore_index:
        if args.csv_no_index:
            log.error("If --ignore-index is set, you must also set --csv-no-index.")
            sys.exit(1)

    # check that uri-root and convert-cols are both present, or neither
    if (args.convert_cols is None and args.uri_root_out is not None) or \
       (args.convert_cols is not None and args.uri_root_out is None):
       log.error("Must specify both --convert-cols and --uri-root-out.")
       sys.exit(1)

# check arguments for concat
def check_concat_arguments(log, args):

    # if origin names provided, check that there are same number of files
    if args.origin_col_names is not None:
        
        # check that user request an origin column
        if args.add_origin_col is None:
            log.error("Must use --add-origin-col if providing origin column names.")
            sys.exit(1)

        # check for matching origin names and files to concatenate
        if len(args.origin_col_names) != len(args.concat):
            log.error("Number of --origin-col-names does not match number of files to concatenate.")
            sys.exit(1)

# check argumetns for expand
def check_expand_arguments(log, args):

    # check that column to expand is provided
    if args.expand_header is None:
        log.error("Please specify --expand-header and try again.")
        sys.exit(1)

# check command arguments
def check_arguments(log, args):
    
    # convert user selection into True/False is option is selected
    options_selected = [vars(args)["create"]] + [vars(args)[option] is not None 
                        for option in ["join", "concat", "expand"]]

    # check that one option is selected
    if sum(options_selected) == 0:
        log.error("Please select one of --create, --join, --concat, and --expand " +
                  "and try again.")
        sys.exit(1)

    # make sure only one option is selected
    if sum(options_selected) > 1:
        log.error("Select only one of --create, --join, --concat, or --expand and try again.")
        sys.exit(1)

    # make sure the output directory is present
    if args.output_dir is None:
        log.error("Output directory must be specified.  Please use --output-dir and try again.")
        sys.exit(1)

    # make sure the output argument is present
    if args.csv_out is None:
        log.error("Name of .csv output file is required.  Please use --csv-out and try again.")
        sys.exit(1)

    # check create options
    if args.create:
        check_create_arguments(log, args)
    
    # check join options
    if args.join is not None:
        check_join_arguments(log, args)
        
    # check concat option
    if args.concat is not None:
        check_concat_arguments(log, args)

    # check expand option
    if args.expand is not None:
        check_expand_arguments(log, args)

# create .csv file
def create_csv(args, log, plugin):

    # create ensemble table
    ensemble_table = EnsembleTable(log, ensemble_spec=args.ensemble, 
        file_spec=args.input_files, header=args.input_header)
    ensemble_dirs = ensemble_table.get_col(args.input_header)

    # quit if no directories found
    num_ensemble_dirs = len(ensemble_dirs)
    if num_ensemble_dirs == 0:
        log.error("No ensemble directories found.  " +
                  "Please identify existing directories and try again.")
        sys.exit(1)
    else:
        log.info("Found %d ensemble directory(ies)." % num_ensemble_dirs)
        
    # go through each directory and read input files
    input_data = []
    for i in range(num_ensemble_dirs):

        # find files in directory
        files_to_read = ensemble_table.files(ensemble_dirs[i])
        log.info("Found %d file(s) in ensemble directory %s." % 
            (len(files_to_read), ensemble_dirs[i]))
        
        # print warning if no files found for input
        if files_to_read == []:
            log.error("No files to read, please provide existing files for input.")
            sys.exit(1)

        # read input files
        input_data.append(plugin.read_input_deck(files_to_read, file_type=args.input_format))

    # combine all input data headers
    input_headers = []
    for i in range(num_ensemble_dirs):
        for key in input_data[i].keys():
            if key not in input_headers:
                input_headers.append(key)
    
    # create table using input data headers
    input_table = []
    for header in input_headers:

        # create column for a given header
        input_col = []
        for i in range(num_ensemble_dirs):
            if header in input_data[i]:
                input_col.append(input_data[i][header])
            else:

                # use empty strings if no data
                input_col.append('')
        
        # check if we should add this column
        if args.csv_headers is not None:
            if len(args.csv_headers) > 0:
                if header not in args.csv_headers:
                    continue

        # check if we should exclude this column
        if args.exclude_csv_headers is not None:
            print("exclude")
            print(args)
            if len(args.csv_headers) > 0:
                if header in args.exclude_csv_headers:
                    continue

        # add column to table
        ensemble_table.add_col(input_col, header)
    
    # write out table
    csv_out = os.path.join(args.output_dir, args.csv_out)
    ensemble_table.to_csv(csv_out, index=args.csv_no_index, 
                          index_label=args.csv_index_header, 
                          cols=args.csv_headers,
                          exc_cols=args.exclude_csv_headers)

# join csv files
def join_csv(args, log):

    # create ensemble table for each .csv file
    ensemble_tables = []
    for csv_file in args.join:
        ensemble_tables.append(EnsembleTable(log, csv_file=csv_file))

    # create extra column if user is using --ensemble
    if args.ensemble is not None:
        ensemble_tables.append(EnsembleTable(log, ensemble_spec=args.ensemble, 
            file_spec=args.input_files, header=args.input_header))
    
    # combine tables
    combined_table = ensemble_combine(log, ensemble_tables, ignore_index=args.ignore_index)
    
    # convert file pointers, if requested
    if args.convert_cols is not None:
        combined_table.convert_cols(args.convert_cols, args.uri_root_out)

    # write out combined table
    csv_out = os.path.join(args.output_dir, args.csv_out)
    combined_table.to_csv(csv_out, index=args.csv_no_index, 
                          index_label=args.csv_index_header,
                          cols=args.csv_headers, 
                          exc_cols=args.exclude_csv_headers)

# concat csv files
def concat_csv(args, log):

    # create ensemble table for each .csv file
    ensemble_tables = []
    for csv_file in args.concat:
        ensemble_tables.append(EnsembleTable(log, csv_file=csv_file))

    # check that column headers are identical
    headers = list(ensemble_tables[0].table)
    headers_identical = True
    for table in ensemble_tables:
        if list(table.table) != headers:
            headers_identical = False

    # quit if headers are not identical
    if not headers_identical:
        log.error("Table headers are not identical, cannot concatenate tables.")
        sys.exit(1)

    # concatenate tables
    concat_table = ensemble.concat(log, ensemble_tables)

    # create origin column, if requested
    origin_col = []
    for i in range(len(ensemble_tables)):

        # use given origin names if provided
        if args.origin_col_names is not None:
            origin_col += [args.origin_col_names[i]] * ensemble_tables[i].table.shape[0]

        # otherwise, use file names
        else:
            origin_col += [args.concat[i]] * ensemble_tables[i].table.shape[0]

    # add origin column to concatenated table
    concat_table.add_col(origin_col, args.add_origin_col)

    # write out new table
    csv_out = os.path.join(args.output_dir, args.csv_out)
    concat_table.to_csv(csv_out, index=args.csv_no_index, 
                          index_label=args.csv_index_header,
                          cols=args.csv_headers,
                          exc_cols=args.exclude_csv_headers)
    

# expand csv file
def expand_csv(args, log, plugin):

    # read main table
    table_to_expand = EnsembleTable(log, csv_file=args.expand)

    # check if column exists
    col_to_expand = table_to_expand.get_col(args.expand_header)
    
    # get files to expand, count files per specifier
    files_to_expand = []
    missing_files = False
    multiple_files = False
    for file_spec in col_to_expand:

        # get files for specifier
        expand_files = table_to_expand.files(file_spec)

        # do the references files exist?
        if len(expand_files) == 0:
            missing_files = True

        # do the references point to multiple files?
        if len(expand_files) > 1:
            multiple_files = True

        # if only one file, make into a string instead of a list
        if len(expand_files) == 1:
            expand_files = expand_files[0]
        
        files_to_expand.append(expand_files)
    
    # if there are missing files, error out
    if missing_files:
        log.error("Column to expand does not reference existing files.")
        sys.exit(1)

    # for multiple files, expand columns but do not read files
    if multiple_files:

        # explode table
        exploded_table = ensemble.explode(log, table_to_expand, 
            args.expand_header, files_to_expand)

        # write out table
        csv_out = os.path.join(args.output_dir, args.csv_out)
        exploded_table.to_csv(csv_out, index=args.csv_no_index, 
                              index_label=args.csv_index_header,
                              cols=args.csv_headers,
                              exc_cols=args.exclude_csv_headers)

    # otherwise use plugin to read files and expand
    # (plugin can also write additional files, if desired)
    else:

        plugin.expand(table_to_expand, args.expand_header, files_to_expand, 
            output_dir=args.output_dir, csv_out=args.csv_out, csv_no_index=args.csv_no_index, 
            csv_index_header=args.csv_index_header, csv_headers=args.csv_headers)

# creates a .csv file for remaining ensemble tools to use as input
# call from Python using arg_list
def main(arg_list=None):

    # initialize parser 
    parser = init_parser()

    # parse arguments
    if arg_list is not None:

        # parse in case called with arg_list
        args, arg_list = parser.parse_args(arg_list)

    else:

        # command line parser
        args, arg_list = parser.parse_args()

    # start logger
    ensemble.init_logger(log_file=args.log_file, log_level=args.log_level)
    log = logging.getLogger("ensemble.table")
    log.debug("Started table.")

    # check arguments
    check_arguments(log, args)
    
    # import and initialize plugin
    plugin, unknown_args = ensemble.plugin(args.plugin, arg_list)

    # check for un-recognized arguments
    if len(unknown_args) > 0:
        log.error("Unrecognized arguments: %s.  Please try again." % str(unknown_args))
        sys.exit(1)
    
    # check if output directory exists
    if not os.path.exists(args.output_dir):
        log.warning("Output directory does not exist -- creating directory: " + 
                    args.output_dir)
        os.makedirs(args.output_dir)

    # check if output file exists
    csv_out = os.path.join(args.output_dir, args.csv_out)
    if os.path.isfile(csv_out):
        if not args.over_write:
            log.error("Output file already exists, use --over-write if you " +
                "want to over-write file.")
            sys.exit(1)

    # create csv
    if args.create:
        create_csv(args, log, plugin)
    
    # join csv
    elif args.join is not None:
        join_csv(args, log)

    # concatenate csv
    elif args.concat is not None:
        concat_csv(args, log)

    # expand csv
    elif args.expand is not None:
        expand_csv(args, log, plugin)

# entry point for command line call
if __name__ == "__main__":
    main()
